r"""
Random variables on a finite probability spaces

This introduces a class of random variables, with the focus on
discrete random variables on a finite probability space.


(C) 2007 David Kohel 
(C) 2012 Tvrtko Tadic
"""

#*****************************************************************************
#       Copyright (C) 2007 David Kohel <kohel@maths.usyd.edu.au>
#       Copyright (C) 2012 Tvrtko Tadic <tadic@uw.edu>
#		
#  Distributed under the terms of the GNU General Public License (GPL)
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.rings.ring import Ring
from sage.misc.functional import sqrt, log
from sage.rings.all import RealField, is_RealField, is_RationalField 
from sage.rings.integer_ring import ZZ
from sage.structure.element import RingElement
from sage.symbolic.ring import SR
from sage.sets.set import Set

################################################################################
################################################################################

def is_FiniteProbabilitySpace_generic(S):
    return isinstance(S, FiniteProbabilitySpace_generic)

def is_FiniteProbabilitySpace(S):
    return isinstance(S, FiniteProbabilitySpace)

def is_RandomVariable_ring(X):
    return isinstance(X, RandomVariable_generic)

def is_DiscreteRandomVariable(X):
    return isinstance(X, DiscreteRandomVariable)

################################################################################
################################################################################

class RandomVariable_ring(Ring):
  """
  Creates a ring of random variables on a probability space.

  R=RandomVariable_ring(PS)

  INPUT:

  PS - finite probability space

  The base of the ring is given by the argument PS._RV_codomain. 


  """

  def __init__(self,PS):
	"""
	Creates a ring of random variables on a probability space.

	R=RandomVariable_ring(PS)

	INPUT:

	PS - finite probability space

	The base of the ring is given by the argument PS._RV_codomain. 


	"""
	if not is_FiniteProbabilitySpace(PS):
		raise TypeError, "Argument is %s not a FiniteProbabilitySpace_generic" % PS
	
	Ring.__init__(self,base=PS._RV_codomain)
	self._PS=PS
	self._codomain=PS._RV_codomain
  def _repr_(self):
	return "Random variables on the given sample space with probability"
  def _element_constructor_(self, x):
        """
        Make sure x is a valid member of self, and return the 	constructed element.
        """
	if isinstance(x, RandomVariable):
		x=x._function;
	else:
		x=self._codomain(x)

	return RandomVariable( x, self._PS, self);
  def _coerce_map_from_(self, S):
	"""
        This map insures that we can add constants to random variables in the ring.
        """

	if (S is SR) or is_RealField(S) or is_RationalField(S) or (S is ZZ) :
		return True
        elif isinstance(S, RandomVariable_ring):
		if S._PS==self._PS:
                	return True
        else:
                return False


class RandomVariable(RingElement):
   """
   This creates a random variable with the function given in the rule,
   on a FiniteProbabilitySpace PS, in the ring parent. If parent ==None
   we will take the ring of the random variables, already defined in the probability
   space.
		
	X=RandomVariable(rule,PS, parent=None)

	INPUT:

	rule - function given by a dictionary on PS._domain

	PS - finite probability space

	parent - the ring of random variables, if None the ring of random variables of PS (PS._RandomVariableRing) will be assigned

	  EXAMPLES::

	     sage: dice={}
	     sage: for k in range(1,7): dice[k]=1/6
	     sage: dice=FiniteProbabilitySpace(dice)
	     sage: #even is the radnom variable that returns 1 if the number on the dice is even, 0 otherwise
	     sage: even=dict([(k,1-k%2) for k in dice.domain()]) 
	     sage: even=RandomVariable(even,dice)



   """
   def __init__(self, rule,PS, parent=None):
	"""
	X=RandomVariable(rule,PS, parent=None)

	"""
	if parent==None:
		parent=PS._RandomVariableRing
   	RingElement.__init__(self, parent)
	self._codomain=PS._RV_codomain;
	if isinstance(rule, dict):
   		self._function = rule
	else:
		x=self._codomain(rule)
		self._function = dict([(y,x) for y in PS._domain])
	self._PS=PS
   def _repr_(self):
        return "Random variable %s on %s" % (self._function,self._PS)

   def _add_(left, right):
	d= dict([(x,left._function[x] + right._function[x]) for x in left._PS._domain])
	return RandomVariable( d, left._PS,left.parent())

   def _sub_(left, right):
	d= dict([(x,left._function[x] - right._function[x]) for x in left._PS._domain])
	return RandomVariable( d, left._PS,left.parent())

   def _mul_(left, right):
	d= dict([(x,left._function[x] * right._function[x]) for x in left._PS._domain])
	return RandomVariable( d, left._PS,left.parent())

   def _div_(left, right):
	if 0 in right._function.values():
			raise ZeroDivisionError, \
		         	"Argument right (= %s) obtains a zero value!" % right
	d= dict([(x,left._function[x] / right._function[x]) for x in left._PS._domain])
	return RandomVariable( d, left._PS,left.parent())


   def _rmul_(self, c):
       d= dict([(x,self._function[x] * c) for x in self._PS._domain])
       return RandomVariable( d, self._PS,self.parent())

   def _lmul_(self, c):
       d= dict([(x,self._function[x] * c) for x in self._PS._domain])
       return RandomVariable( d, self._PS,self.parent())

   def equals(self,right):
	r"""
	X.equals(Y)
	This function returns the set `\{\omega\in \Omega: X(\omega)=Y(\omega)\}` on which
	the random variables X and Y are equal. Y can be an element of the codomain. 

	  EXAMPLES::

	     sage: dice={}
	     sage: for k in range(1,7): dice[k]=1/6
	     sage: dice=FiniteProbabilitySpace(dice)
	     sage: #even is the radnom variable that returns 1 if the number on the dice is even, 0 otherwise
	     sage: even=dict([(k,1-k%2) for k in dice.domain()]) 
	     sage: even=FiniteRandomVariable(even,dice)
	     sage: # the random variable that returns 1 if the dice shows 6, 0 if not
	     sage: six=dict([(k,1) for k in dice.domain() if k==6]+[(k,0) for k in dice.domain() if k!=6]) 
	     sage: six=FiniteRandomVariable(six,dice)
	     sage: six.equals(0)
	     set([1,2,3,4,5])
	     sage: six.equals(even)
	     set([1,3,5,6])	
	"""
	if self.parent()==right.parent():
		return set([a for a in self._PS._domain if self(a)==right(a)])
	else:
		x=self._codomain(right)
		return set([a for a in self._PS._domain if self(a)==x])
   def in_interval(self,x,y):
	r"""
	This function returns the set `\{\omega\in \Omega: Y(\omega)\leq X(\omega)\leq Z(\omega)\}` on which
	the random variable X is in the interval [Y,Z]. Y,Z can be an element of the codomain or random variables. 

	X.in_interval(Y,Z)

	INPUT:

	Y,Z - random variables in same parent RandomVariableRing as X or constants in the base of the parent


  	  EXAMPLES::

	     sage: dice={}
	     sage: for k in range(1,7): dice[k]=1/6
	     sage: dice=FiniteProbabilitySpace(dice)
	     sage: #even is the radnom variable that returns 1 if the number on the dice is even, 0 otherwise
	     sage: even=dict([(k,1-k%2) for k in dice.domain()]) 
	     sage: even=FiniteRandomVariable(even,dice)
	     sage: # the random variable that returns 1 if the dice shows 6, 0 if not
	     sage: six=dict([(k,1) for k in dice.domain() if k==6]+[(k,0) for k in dice.domain() if k!=6]) 
	     sage: six=FiniteRandomVariable(six,dice)
	     sage: six.in_interval(even,6)
	     set([1, 3, 5, 6])

	"""
	if self.parent()==x.parent() and self.parent()==y.parent():
		return set([a for a in self._PS._domain if (self(a)<=y(a)) and (x(a)<=self(a))])		
	elif self.parent()!=x.parent() and self.parent()==y.parent():
		x=self._codomain(x);
		return set([a for a in self._PS._domain if (self(a)<=y(a)) and (x<=self(a))])
	elif self.parent()==x.parent() and self.parent()!=y.parent():
		y=self._codomain(y);
		return set([a for a in self._PS._domain if (self(a)<=y) and (x(a)<=self(a))])
	else:
		x=self._codomain(x);	y=self._codomain(y);
		return set([a for a in self._PS._domain if (self(a)<=y) and (x<=self(a))])

   def __call__(self,x):
        r"""
        Return the value of the random variable at x.

	  EXAMPLE::

	     sage: dice={}
	     sage: for k in range(1,7): dice[k]=1/6
	     sage: dice=FiniteProbabilitySpace(dice)
	     sage: #even is the radnom variable that returns 1 if the number on the dice is even, 0 otherwise
	     sage: even=dict([(k,1-k%2) for k in dice.domain()]) 
	     sage: even=FiniteRandomVariable(even,dice)
	     sage: even(3), even(2), even(4)
	     (0, 1, 1)

        """
        
        try:
            return self._function[x]
        except KeyError:
            # Need some condition for x being a valid domain element:
                raise IndexError, "Argument x (= %s) is not a valid domain element." % x
            	#return self._codomain(0)
   def function(self):
        r"""
        The function defining the random variable.
	Avoid using for large probability spaces.

	  EXAMPLES::

	     sage: dice={}
	     sage: for k in range(1,7): dice[k]=1/6
	     sage: dice=FiniteProbabilitySpace(dice)
	     sage: #even is the radnom variable that returns 1 if the number on the dice is even, 0 otherwise
	     sage: even=dict([(k,1-k%2) for k in dice.domain()]) 
	     sage: even=FiniteRandomVariable(even,dice)
	     {1: 0, 2: 1, 3: 0, 4: 1, 5: 0, 6: 1} 
        """
        return self._function
   def probability_space(self):

	return self._PS

   def expectation(self):
        r"""
        The expectation of the discrete random variable, namely
        `\sum_{x \in S} p(x) X[x]`, where `X` = self and
        `S` is the probability space of `X`.

	  EXAMPLES:
	     sage: dice={}
	     sage: for k in range(1,7): dice[k]=1/6
	     sage: dice=FiniteProbabilitySpace(dice)
	     sage: #even is the radnom variable that returns 1 if the number on the dice is even, 0 otherwise
	     sage: even=dict([(k,1-k%2) for k in dice.domain()]) 
	     sage: even=FiniteRandomVariable(even,dice)
	     sage: # the random variable that returns 1 if the dice shows 6, 0 if not
	     sage: six=dict([(k,1) for k in dice.domain() if k==6]+[(k,0) for k in dice.domain() if k!=6]) 
	     sage: six=FiniteRandomVariable(six,dice)
	     sage: even.expectation()
	     1/2
	     sage: six.expectation()
	     1/6
	     sage: six.expectation()
	     1/6
	     sage: (six*even).expectation()
	     1/6


        """
        E = 0
        Omega = self.probability_space()
        for x in Omega._domain:
            E += Omega._probability[x] * self(x)
        return E

   def translation_expectation(self, map):
        r"""
        The expectation of the discrete random variable, namely
        `\sum_{x \in S} p(x) X[e(x)]`, where `X` = self,
        `S` is the probability space of `X`, and
        `e` = map.


	  EXAMPLES:
	     sage: dice={}
	     sage: for k in range(1,7): dice[k]=1/6
	     sage: dice=FiniteProbabilitySpace(dice)
	     sage: #even is the radnom variable that returns 1 if the number on the dice is even, 0 otherwise
	     sage: even=dict([(k,1-k%2) for k in dice.domain()]) 
	     sage: even=FiniteRandomVariable(even,dice)
	     sage: f(x)=7-x
	     sage: even.translation_expectation(f)
	     1/2
        """
        E = 0
        Omega = self.probability_space()
        for x in Omega._domain:
            E += Omega._probability[x] * self(map(x))
        return E

   def variance(self):
        r"""
        The variance of the discrete random variable.
        
        Let `S` be the probability space of `X` = self,
        with probability function `p`, and `E(X)` be the
        expectation of `X`. Then the variance of `X` is:
        
        .. math::
        
           \mathrm{var}(X) = E((X-E(x))^2) = \sum_{x \in S} p(x) (X(x) - E(x))^2

	  EXAMPLES::

	     sage: dice={}
	     sage: for k in range(1,7): dice[k]=1/6
	     sage: dice=FiniteProbabilitySpace(dice)
	     sage: #even is the radnom variable that returns 1 if the number on the dice is even, 0 otherwise
	     sage: even=dict([(k,1-k%2) for k in dice.domain()]) 
	     sage: even=FiniteRandomVariable(even,dice)
	     sage: # the random variable that returns 1 if the dice shows 6, 0 if not
	     sage: six=dict([(k,1) for k in dice.domain() if k==6]+[(k,0) for k in dice.domain() if k!=6]) 
	     sage: six=FiniteRandomVariable(six,dice)
	     sage: even.variance()
	     1/4
	     sage: six.variance()
	     5/36
	     sage: (six*even).variance()
	     5/36

        """
        Omega = self.probability_space()
        mu = self.expectation()
        var = - mu*mu   
        for x in Omega._domain:
            var += Omega._probability[x] * (self(x)**2)
        return var

   def translation_variance(self, map):
        r"""
        The variance of the discrete random variable `X \circ e`,
        where `X` = self, and `e` = map.
        
        Let `S` be the probability space of `X` = self,
        with probability function `p`, and `E(X)` be the
        expectation of `X`. Then the variance of `X` is:
        
        .. math::
        
           \mathrm{var}(X) = E((X\circ e-E(X\circ e))^2) = \sum_{x \in S} p(x) (X(e(x)) - E(e(x)))^2

	  EXAMPLES:
	     sage: dice={}
	     sage: for k in range(1,7): dice[k]=1/6
	     sage: dice=FiniteProbabilitySpace(dice)
	     sage: #even is the radnom variable that returns 1 if the number on the dice is even, 0 otherwise
	     sage: even=dict([(k,1-k%2) for k in dice.domain()]) 
	     sage: even=FiniteRandomVariable(even,dice)
	     sage: f(x)=7-x
	     sage: even.translation_variance(f)
	     1/4

        """
        Omega = self.probability_space()
        mu = self.translation_expectation(map)
        var =  - mu*mu
        for x in Omega._domain:
            var += Omega._probability[x] * (self(map(x))**2 )
        return var

   def covariance(self, other):
        r"""
        The covariance of the discrete random variable X = self with Y =
        other.
        
        Let `S` be the probability space of `X` = self,
        with probability function `p`, and `E(X)` be the
        expectation of `X`. Then the variance of `X` is:
        
        .. math::
        
                     \text{cov}(X,Y) = E((X-E(X)*(Y-E(Y)) = \sum_{x \in S} p(x) (X(x) - E(X))(Y(x) - E(Y))

	  EXAMPLES:
	     sage: dice={}
	     sage: for k in range(1,7): dice[k]=1/6
	     sage: dice=FiniteProbabilitySpace(dice)
	     sage: #even is the radnom variable that returns 1 if the number on the dice is even, 0 otherwise
	     sage: even=dict([(k,1-k%2) for k in dice.domain()]) 
	     sage: even=FiniteRandomVariable(even,dice)
	     sage: # the random variable that returns 1 if the dice shows 6, 0 if not
	     sage: six=dict([(k,1) for k in dice.domain() if k==6]+[(k,0) for k in dice.domain() if k!=6]) 
	     sage: six=FiniteRandomVariable(six,dice)
	     sage: six.covariance(even)
	     1/12
	     sage: six.covariance(six) #should be the same as six.variance()
	     5/36

        """
        Omega = self.probability_space()
        if Omega != other.probability_space():
            raise ValueError, \
                 "Argument other (= %s) must be defined on the same probability space." % other
        muX = self.expectation()
        muY = other.expectation()
        cov = -muX*muY
        for x in Omega._domain:
            cov += Omega._probability[x]*(self(x))*(other(x))
        return cov

   def translation_covariance(self, other, map):
        r"""
        The covariance of the probability space X = self with image of Y =
        other under the given map e of the probability space.
        
        Let `S` be the probability space of `X` = self,
        with probability function `p`, and `E(X)` be the
        expectation of `X`. Then the variance of `X` is:
        
        .. math::
        
                     \text{cov}(X,Y) = E((X-E(X)*(Y\circ e-E(Y\circ e)) = \sum_{x \in S} p(x) (X(x) - E(X))(Y(e(x)) - E(Y\circ e))

	  EXAMPLES::

	     sage: dice={}
	     sage: for k in range(1,7): dice[k]=1/6
	     sage: dice=FiniteProbabilitySpace(dice)
	     sage: #even is the radnom variable that returns 1 if the number on the dice is even, 0 otherwise
	     sage: even=dict([(k,1-k%2) for k in dice.domain()]) 
	     sage: even=FiniteRandomVariable(even,dice)
	     sage: # the random variable that returns 1 if the dice shows 6, 0 if not
	     sage: six=dict([(k,1) for k in dice.domain() if k==6]+[(k,0) for k in dice.domain() if k!=6]) 
	     sage: six=FiniteRandomVariable(six,dice)
	     sage: f(x)=7-x
	     sage: six.translation_covariance(even,f)
	     -1/12


        """
        Omega = self.probability_space()
        if Omega != other.probability_space():
            raise ValueError, \
                 "Argument other (= %s) must be defined on the same probability space." % other
        muX = self.expectation()
        muY = other.translation_expectation(map)
        cov = -muX*muY
        for x in Omega._domain:
            cov += Omega._probability[x]*(self(x))*(other(map(x)))
        return cov

   def standard_deviation(self):
        r"""
        The standard deviation of the discrete random variable.
        
        Let `S` be the probability space of `X` = self,
        with probability function `p`, and `E(X)` be the
        expectation of `X`. Then the standard deviation of
        `X` is defined to be
        
        .. math::
        
                     \sigma(X) = \sqrt{ \sum_{x \in S} p(x) (X(x) - E(x))**2}

	  EXAMPLES::

	     sage: dice={}
	     sage: for k in range(1,7): dice[k]=1/6
	     sage: dice=FiniteProbabilitySpace(dice)
	     sage: #even is the radnom variable that returns 1 if the number on the dice is even, 0 otherwise
	     sage: even=dict([(k,1-k%2) for k in dice.domain()]) 
	     sage: even=FiniteRandomVariable(even,dice)
	     sage: even.standard_deviation()
	     1/2


        """
        return sqrt(self.variance())

   def translation_standard_deviation(self, map):
        r"""
        The standard deviation of the translated discrete random variable
        `X \circ e`, where `X` = self and `e` =
        map.
        
        Let `S` be the probability space of `X` = self,
        with probability function `p`, and `E(X)` be the
        expectation of `X`. Then the standard deviation of
        `X` is defined to be
        
        .. math::
        
                     \sigma(X) = \sqrt{ \sum_{x \in S} p(x) (X(x) - E(x))**2}
	  EXAMPLES::

	     sage: dice={}
	     sage: for k in range(1,7): dice[k]=1/6
	     sage: dice=FiniteProbabilitySpace(dice)
	     sage: #even is the radnom variable that returns 1 if the number on the dice is even, 0 otherwise
	     sage: even=dict([(k,1-k%2) for k in dice.domain()]) 
	     sage: even=FiniteRandomVariable(even,dice)
	     sage: f(x)=7-x
	     sage: even.translation_standard_deviation(f)
	     1/2


        """
        return sqrt(self.translation_variance(map))

   def correlation(self, other):
        r"""
        The correlation of the probability space X = self with Y = other.

	  EXAMPLES::

	     sage: dice={}
	     sage: for k in range(1,7): dice[k]=1/6
	     sage: dice=FiniteProbabilitySpace(dice)
	     sage: #even is the radnom variable that returns 1 if the number on the dice is even, 0 otherwise
	     sage: even=dict([(k,1-k%2) for k in dice.domain()]) 
	     sage: even=FiniteRandomVariable(even,dice)
	     sage: # the random variable that returns 1 if the dice shows 6, 0 if not
	     sage: six=dict([(k,1) for k in dice.domain() if k==6]+[(k,0) for k in dice.domain() if k!=6]) 
	     sage: six=FiniteRandomVariable(six,dice)
	     sage: six.correlation(even)
	     1/5*sqrt(5)
	     sage: six.correlation(6*six+1) #should be 1
	     1
        """
        cov = self.covariance(other)
        sigX = self.standard_deviation()
        sigY = other.standard_deviation()
        if sigX == 0 or sigY == 0:
            raise ValueError, \
                "Correlation not defined if standard deviations are not both nonzero."
        return cov/(sigX*sigY)

   def translation_correlation(self, other, map):
        """
        The correlation of the probability space X = self with image of Y =
        other under map.

	  EXAMPLES::

	     sage: dice={}
	     sage: for k in range(1,7): dice[k]=1/6
	     sage: dice=FiniteProbabilitySpace(dice)
	     sage: #even is the radnom variable that returns 1 if the number on the dice is even, 0 otherwise
	     sage: even=dict([(k,1-k%2) for k in dice.domain()]) 
	     sage: even=FiniteRandomVariable(even,dice)
	     sage: # the random variable that returns 1 if the dice shows 6, 0 if not
	     sage: six=dict([(k,1) for k in dice.domain() if k==6]+[(k,0) for k in dice.domain() if k!=6]) 
	     sage: six=FiniteRandomVariable(six,dice)
	     sage: f(x)=7-x
	     sage: six.translation_correlation(even,f)
	     -1/5*sqrt(5)

        """
        cov = self.translation_covariance(other, map)
        sigX = self.standard_deviation()
        sigY = other.translation_standard_deviation(map)
        if sigX == 0 or sigY == 0:
            raise ValueError, \
                "Correlation not defined if standard deviations are not both nonzero."
        return cov/(sigX*sigY)

class FiniteRandomVariable(RandomVariable):
    r"""
    A random variable on a finite probability space.
		
	  EXAMPLES::

	     sage: dice={}
	     sage: for k in range(1,7): dice[k]=1/6
	     sage: dice=FiniteProbabilitySpace(dice)
	     sage: #even is the radnom variable that returns 1 if the number on the dice is even, 0 otherwise
	     sage: even=dict([(k,1-k%2) for k in dice.domain()]) 
	     sage: even=FiniteRandomVariable(even,dice)
	     sage: # the random variable that returns 1 if the dice shows 6, 0 if not
	     sage: six=dict([(k,1) for k in dice.domain() if k==6]+[(k,0) for k in dice.domain() if k!=6]) 
	     sage: six=FiniteRandomVariable(six,dice)


    """
    def __init__(self, f,PS):
        r"""
	X=FiniteRandomVariable(f,PS)

        INPUT: A probability space PS. A dictionary such that f[x] =
        value for x in PS._domain.
        """
        if not is_FiniteProbabilitySpace(PS):
            raise TypeError, "Argument PS (= %s) must be a discrete probability space" % PS
        RandomVariable.__init__(self, f, PS)



class DiscreteRandomVariable(RandomVariable):
    """
    A random variable on a discrete probability space.
    This is an implementation of the class DicreteRandomVariable that was used in previous versions of SAGE.

    X=FiniteRandomVariable(PS,f,codomain=None,check=False)

    INPUT: A probability space PS. A dictionary such that f[x] =
    value for x in PS._domain.

	EXAMPLES::

	    sage: AZ = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
            sage: S = [ AZ[i] for i in range(26) ]
            sage: P = { 'A':1/2, 'B':1/4, 'C':1/4 }
            sage: X = DiscreteProbabilitySpace(S,P)
            sage: from sage.crypto.util import ascii_integer
	    sage: bin = BinaryStrings()
	    sage: Y=dict([(a,ascii_integer(bin.encoding(a))) for a in S]);
	    sage: Y=DiscreteRandomVariable(X,Y);
	    
    """
    def __init__(self,PS,f,codomain=None,check=False):
        r"""
	X=FiniteRandomVariable(f,PS)
        INPUT: A probability space PS. A dictionary such that f[x] =
        value for x in PS._domain.
        """
        if not isinstance(PS,DiscreteProbabilitySpace):
            raise TypeError, "Argument PS (= %s) must be a DiscreteProbabilitySpace" % PS
        RandomVariable.__init__(self, f, PS)

    

#######################################################################################
#######################################################################################

class FiniteProbabilitySpace():
    r"""
    The finite probability space.

    FiniteProbabilitySpace(P, codomain = None, check = False,RV_codomain=None)

    P - dictionary containing the outcomes in keys and their probabilities in values

    codomain - the ring in which the probabilities are given RR/QQ/SR (if None codomain=SR)

    RV_codomain  - the ring in which the random variables have values RR/QQ/SR (if None RV_codomain=SR)

    check - see if the probabilities sum to 1

        EXAMPLES::

            sage: dice={}
	    sage: for(k in range(1,7)): dice[k]=1/6
	    sage: dice=FiniteProbabilitySpace(dice,codomain=QQ)
	    sage: dice
	    Discrete probability space defined by {0: 1/6, 1: 1/6, 2: 1/6, 3: 1/6, 4: 1/6, 5: 1/6, 6: 1/6}
	    sage: dice(4)
	    1/6
	    sage: dice([1,2,3])
	    1/2        
            sage: dice.codomain()
	    Rational Field
	    sage: dice.set()
	    set([1, 2, 3, 4, 5, 6])
	    sage: dice.entropy()
            2.58496250072

        EXAMPLES::
        
            sage: Omega=['W','L'] #sample space
            sage: p=var('p'); P=[p,1-p] #symbolic probabilities
            sage: Pr=dict([(Omega[k],P[k]) for k in range(0,len(Omega))]) #probability function
            sage: PS=FiniteProbabilitySpace(Pr)
            sage: PS
            Finite probability space defined by {'L': -p + 1, 'W': p}
            sage: PS('W') #probability of 'W'
	    p
	    sage: PS(['W','L']) #probability of 'W' or 'L'
	    1
                   


    """
    def __init__(self, P, codomain = None, check = False, RV_codomain=None):
        r"""
        Create the finite probability space with probabilities on the
        space X given by the dictionary P with values in the codomain.
        
        """

        if codomain is None:
           codomain = SR
	if RV_codomain is None:
           RV_codomain = SR

        if not is_RealField(codomain) and not is_RationalField(codomain) and not (codomain is SR):
            raise TypeError, "Argument codomain (= %s) must be the reals or rationals or symbols" % codomain
        if not is_RealField(RV_codomain) and not is_RationalField(RV_codomain) and not (RV_codomain is SR):
            raise TypeError, "Argument codomain (= %s) must be the reals or rationals or symbols" % codomain

        if check:
	    if not all( [x in codomain for x in P.values()]) :
		    raise TypeError, "Probabilities must be in codomain(= %s)" % codomain

            one = sum([ P[x] for x in P.keys() ])
            if is_RationalField(codomain):    
                if not one == 1:
                    raise TypeError, "Argument P (= %s) does not define a probability function"
            elif is_RealField(codomain):
                if not Abs(one-1) < 2^(-codomain.precision()+1):
                    raise TypeError, "Argument P (= %s) does not define a probability function"
	    else:
		print("Check if")
		print one
		print("equals 1")		
	self._domain = P.keys();
	self._probability=P;
	self._codomain=codomain;
	self._RV_codomain=RV_codomain;
	self._RandomVariableRing=RandomVariable_ring(self)
    def __call__(self,x):
        r"""
        Return the probability of the of an event

          EXAMPLES::

             sage: dice={}
  	     sage: for(k in range(1,7)): dice[k]=1/6
	     sage: dice=FiniteProbabilitySpace(dice,codomain=QQ)
	     sage: dice(4)
	     1/6
	     sage: dice([1,2,3])
	     1/2        

        """
	
	try:	
		
		if x in self._domain:
				return self._probability[x]; 
		else:
				return sum([self._probability[y] for y in set(x)])		
	except KeyError,TypeError:
		raise TypeError, "The probability of the argument needs to be an element of the sample space %s or a collection of elements of it." %self.set()
    def domain(self):
	r"""
	Returns the domain of the probability space

           EXAMPLES::

	    sage: dice={}
	    sage: for(k in range(1,7)): dice[k]=1/6
	    sage: dice=FiniteProbabilitySpace(dice,codomain=QQ)
	    sage: dice.domain()
	    [1, 2, 3, 4, 5, 6]
	"""
        return self._domain
    def probability(self):
	r"""
	Shows the probability law on the domain

       	  EXAMPLES::

	    sage: dice={}
	    sage: for(k in range(1,7)): dice[k]=1/6
	    sage: dice=FiniteProbabilitySpace(dice,codomain=QQ)
	    sage: dice.probability()
	    {1: 1/6, 2: 1/6, 3: 1/6, 4: 1/6, 5: 1/6, 6: 1/6}
	"""
	return self._probability
    def codomain(self):
	r"""
	The probabilities in the law are given in this ring.

	  EXAMPLES::

            sage: dice={}
	    sage: for(k in range(1,7)): dice[k]=1/6
	    sage: dice=FiniteProbabilitySpace(dice,codomain=QQ)
            sage: dice.codomain()
	    Rational Field

	"""
        return self._codomain

    def __repr__(self):
	r"""
	The representation of the probability space.
	If the sample space is too large, it you won't be able to see everything.

	  EXAMPLES::

            sage: dice={}
	    sage: for(k in range(1,7)): dice[k]=1/6
	    sage: dice=FiniteProbabilitySpace(dice,codomain=QQ)
	    sage: dice
	    Discrete probability space defined by {0: 1/6, 1: 1/6, 2: 1/6, 3: 1/6, 4: 1/6, 5: 1/6, 6: 1/6}

	"""
        return "Discrete probability space defined by %s" % self._probability
	

        
    def set(self):
        r"""
        The sample space.

	  EXAMPLES::

            sage: dice={}
	    sage: for(k in range(1,7)): dice[k]=1/6
	    sage: dice=FiniteProbabilitySpace(dice,codomain=QQ)
	    sage: dice.set()
	    set([1,2,3,4,5,6])
        """
        return set(self._domain);

    def entropy(self):
        r"""
        The entropy of the probability space.

	  EXAMPLES::

            sage: dice={}
	    sage: for(k in range(1,7)): dice[k]=1/6
	    sage: dice=FiniteProbabilitySpace(dice,codomain=QQ)
	    sage: dice.entropy()
            2.58496250072
        """
        def neg_xlog2x(p):
            if p == 0:
                return 0
            else:  
                return -p*log(p,2)
        p = self._probability
        return sum([ neg_xlog2x(p[x]) for x in p.keys() ])

class DiscreteProbabilitySpace(FiniteProbabilitySpace):
    r"""
    This is an old class that was defined in the previous versions of SAGE.
    Now it is implemented over the class FiniteProbabilitySpace and most
    of the functions that worked in the previous implementation still work.
    
    Create the discrete probability space with probabilities on the
    space X given by the dictionary P with values in the field codomain QQ/RR/SR.
    

    DPS=DiscreteProbabilitySpace(X, P, codomain = None, check = False)
	
	INPUT:

	X the domain

	P the dictionary defining the probability law on a subset of X

	codomain  the ring in which probabilities have values QQ/RR/SR (if None, SR)

	check - check if the sum of probabilities is 1
  
        EXAMPLES::
        
            sage: S = [ i for i in range(16) ] 
            sage: P = {}
                   sage: for i in range(15): P[i] = 2^(-i-1)
            sage: P[15] = 2^-16 
            sage: X = DiscreteProbabilitySpace(S,P)

	EXAMPLES::
        
            sage: AZ = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
            sage: S = [ AZ[i] for i in range(26) ]
            sage: P = { 'A':1/2, 'B':1/4, 'C':1/4 }
            sage: X = DiscreteProbabilitySpace(S,P)
            sage: X
            Discrete probability space defined by {'A': 1/2, 'C': 1/4, 'B': 1/4}
            sage: X.entropy()
                   1.5
    """
    def __init__(self, X, P, codomain = None, check = False):
	""" 
	Initialization of a dicrete probability space

        EXAMPLES::
        
            sage: S = [ i for i in range(16) ] 
            sage: P = {}
                   sage: for i in range(15): P[i] = 2^(-i-1)
            sage: P[15] = 2^-16 
            sage: X = DiscreteProbabilitySpace(S,P)

	"""
	PP=P;
	for x in X:
		if x not in PP.keys():
			PP[x]=0;
	FiniteProbabilitySpace.__init__(self, PP, codomain, check)
